(function (global, factory) {
  if (typeof define === "function" && define.amd) {
    define(['exports', './Nymph', './Entity'], factory);
  } else if (typeof exports !== "undefined") {
    factory(exports, require('./Nymph'), require('./Entity'));
  } else {
    var mod = {
      exports: {}
    };
    factory(mod.exports, global.Nymph, global.Entity);
    global.PubSub = mod.exports;
  }
})(this, function (exports, _Nymph, _Entity) {
  /* global WebSocket */
  'use strict';

  Object.defineProperty(exports, "__esModule", {
    value: true
  });
  exports.PubSubSubscription = exports.PubSub = undefined;

  function _toConsumableArray(arr) {
    if (Array.isArray(arr)) {
      for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) {
        arr2[i] = arr[i];
      }

      return arr2;
    } else {
      return Array.from(arr);
    }
  }

  function _classCallCheck(instance, Constructor) {
    if (!(instance instanceof Constructor)) {
      throw new TypeError("Cannot call a class as a function");
    }
  }

  var _createClass = function () {
    function defineProperties(target, props) {
      for (var i = 0; i < props.length; i++) {
        var descriptor = props[i];
        descriptor.enumerable = descriptor.enumerable || false;
        descriptor.configurable = true;
        if ("value" in descriptor) descriptor.writable = true;
        Object.defineProperty(target, descriptor.key, descriptor);
      }
    }

    return function (Constructor, protoProps, staticProps) {
      if (protoProps) defineProperties(Constructor.prototype, protoProps);
      if (staticProps) defineProperties(Constructor, staticProps);
      return Constructor;
    };
  }();

  var authToken = null;

  if (typeof WebSocket === 'undefined') {
    throw new Error('Nymph-PubSub requires WebSocket!');
  }

  var PubSub = exports.PubSub = function () {
    function PubSub() {
      _classCallCheck(this, PubSub);
    }

    _createClass(PubSub, null, [{
      key: 'init',

      // === Static Methods ===

      value: function init(NymphOptions) {
        this.pubsubURL = NymphOptions.pubsubURL;
        if (NymphOptions.rateLimit) {
          this.rateLimit = NymphOptions.rateLimit;
        }

        this.connect();

        // Override the original Nymph methods to allow subscriptions.
        var getEntities = _Nymph.Nymph.getEntities;
        var getEntity = _Nymph.Nymph.getEntity;
        var getUID = _Nymph.Nymph.getUID;

        _Nymph.Nymph.getEntities = function (options) {
          for (var _len = arguments.length, selectors = Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
            selectors[_key - 1] = arguments[_key];
          }

          var promise = getEntities.apply(_Nymph.Nymph, [options].concat(selectors));
          promise.query = JSON.stringify([options].concat(selectors));
          promise.subscribe = function (resolve, reject, count) {
            var callbacks = [resolve, reject, count];

            promise.then(resolve, reject);

            PubSub.subscribeQuery(promise.query, callbacks);
            return new PubSubSubscription(promise.query, callbacks, function () {
              PubSub.unsubscribeQuery(promise.query, callbacks);
            });
          };
          return promise;
        };

        _Nymph.Nymph.getEntity = function (options) {
          for (var _len2 = arguments.length, selectors = Array(_len2 > 1 ? _len2 - 1 : 0), _key2 = 1; _key2 < _len2; _key2++) {
            selectors[_key2 - 1] = arguments[_key2];
          }

          var promise = getEntity.apply(_Nymph.Nymph, [options].concat(selectors));
          options.limit = 1;
          promise.query = JSON.stringify([options].concat(selectors));
          promise.subscribe = function (resolve, reject, count) {
            var newResolve = function newResolve(args) {
              if (!args.length) {
                if (resolve) {
                  resolve(null);
                }
              } else {
                if (resolve) {
                  resolve(args[0]);
                }
              }
            };
            var callbacks = [newResolve, reject, count];

            promise.then(resolve, reject);

            PubSub.subscribeQuery(promise.query, callbacks);
            return new PubSubSubscription(promise.query, callbacks, function () {
              PubSub.unsubscribeQuery(promise.query, callbacks);
            });
          };
          return promise;
        };

        _Nymph.Nymph.getUID = function (name) {
          var promise = getUID.apply(_Nymph.Nymph, [name]);
          promise.subscribe = function (resolve, reject, count) {
            var callbacks = [resolve, reject, count];

            promise.then(resolve, reject);

            PubSub.subscribeUID(name, callbacks);
            return {
              unsubscribe: function unsubscribe() {
                PubSub.unsubscribeUID(name, callbacks);
              }
            };
          };
          return promise;
        };

        _Entity.Entity.prototype.subscribe = function (resolve, reject, count) {
          var _this = this;

          if (!this.guid) {
            return false;
          }
          var query = [{ 'class': this.constructor.class, 'limit': 1 }, { type: '&', guid: this.guid }];
          var jsonQuery = JSON.stringify(query);

          var newResolve = function newResolve(args) {
            var myArray = void 0;
            if (Array.isArray(args)) {
              myArray = args;
              if (myArray.length) {
                _this.init(myArray[0]);
              }
            } else {
              myArray = [_this];
              PubSub.updateArray(myArray, args);
            }

            if (!myArray.length) {
              _this.guid = null;
            }

            if (resolve) {
              resolve(_this);
            }
          };
          var callbacks = [newResolve, reject, count];

          PubSub.subscribeQuery(jsonQuery, callbacks);
          return {
            unsubscribe: function unsubscribe() {
              PubSub.unsubscribeQuery(jsonQuery, callbacks);
            }
          };
        };

        return this;
      }
    }, {
      key: 'connect',
      value: function connect() {
        var _this2 = this;

        var timedAttemptConnect = function timedAttemptConnect() {
          // Attempt to connect, wait 5 seconds, then check and attempt again if unsuccessful.
          setTimeout(function () {
            if (_this2.connection.readyState !== WebSocket.OPEN) {
              _this2.connection.close();
              timedAttemptConnect();
            }
          }, 5000);

          _this2.connection = new WebSocket(_this2.pubsubURL);

          _this2.connection.onopen = function () {
            if (typeof console !== 'undefined') {
              console.log('Nymph-PubSub connection established!');
            }
            for (var i = 0; i < _this2.connectCallbacks.length; i++) {
              if (typeof _this2.connectCallbacks[i] !== 'undefined') {
                _this2.connectCallbacks[i]();
              }
            }

            if (authToken !== null) {
              _this2.connection.send(JSON.stringify({
                'action': 'authenticate',
                'token': authToken
              }));
            }

            for (var query in _this2.subscriptions.queries) {
              if (!_this2.subscriptions.queries.hasOwnProperty(query)) {
                continue;
              }
              var count = false;
              for (var callbacks = 0; callbacks < _this2.subscriptions.queries[query].length; callbacks++) {
                if (typeof _this2.subscriptions.queries[query][callbacks][2] !== 'undefined') {
                  count = true;
                  break;
                }
              }
              _this2._subscribeQuery(query, count);
            }

            for (var name in _this2.subscriptions.uids) {
              if (!_this2.subscriptions.uids.hasOwnProperty(name)) {
                continue;
              }
              var _count = false;
              for (var _callbacks = 0; _callbacks < _this2.subscriptions.uids[name].length; _callbacks++) {
                if (typeof _this2.subscriptions.uids[name][_callbacks][2] !== 'undefined') {
                  _count = true;
                  break;
                }
              }
              _this2._subscribeUID(name, _count);
            }

            _this2.connection.onclose = function (e) {
              if (typeof console !== 'undefined') {
                console.log('Nymph-PubSub connection closed: ', e);
              }
              for (var _i = 0; _i < _this2.disconnectCallbacks.length; _i++) {
                if (typeof _this2.disconnectCallbacks[_i] !== 'undefined') {
                  _this2.disconnectCallbacks[_i]();
                }
              }
              if (e.code !== 1000) {
                _this2.connection.close();
                timedAttemptConnect();
              }
            };
          };

          _this2.connection.onmessage = function (e) {
            var func = void 0,
                data = void 0;
            if (!_this2.rateLimit || typeof _this2.debouncers[e.data] === 'undefined') {
              func = function func() {
                data = JSON.parse(e.data);
                var val = null;
                var subs = [];
                if (data.hasOwnProperty('query') && typeof _this2.subscriptions.queries[data.query] !== 'undefined') {
                  subs = [].concat(_toConsumableArray(_this2.subscriptions.queries[data.query]));
                  if (!data.hasOwnProperty('count')) {
                    val = data;
                  }
                } else if (data.hasOwnProperty('uid') && typeof _this2.subscriptions.uids[data.uid] !== 'undefined') {
                  subs = [].concat(_toConsumableArray(_this2.subscriptions.uids[data.uid]));
                  if (!data.hasOwnProperty('count') && (data.event === 'newUID' || data.event === 'setUID')) {
                    val = data.value;
                  }
                }
                if (data.hasOwnProperty('count')) {
                  for (var i = 0; i < subs.length; i++) {
                    if (typeof subs[i][2] !== 'undefined') {
                      subs[i][2](data.count);
                    }
                  }
                } else {
                  for (var _i2 = 0; _i2 < subs.length; _i2++) {
                    if (typeof subs[_i2][0] !== 'undefined') {
                      subs[_i2][0](val);
                    }
                  }
                }
                if (_this2.rateLimit) {
                  delete _this2.debouncers[e.data];
                }
              };
            }
            if (!_this2.rateLimit) {
              func();
              return;
            }
            if (typeof _this2.debouncers[e.data] === 'undefined') {
              _this2.debouncers[e.data] = _this2.debounce(func);
            }
            _this2.debouncers[e.data]();
          };
        };
        timedAttemptConnect();
      }
    }, {
      key: 'debounce',
      value: function debounce(func, immediate) {
        var timeout = void 0;
        var that = this;
        return function () {
          for (var _len3 = arguments.length, args = Array(_len3), _key3 = 0; _key3 < _len3; _key3++) {
            args[_key3] = arguments[_key3];
          }

          var context = this;
          var later = function later() {
            timeout = null;
            if (!immediate) {
              func.apply(context, args);
            }
          };
          var callNow = immediate && !timeout;
          clearTimeout(timeout);
          timeout = setTimeout(later, that.rateLimit);
          if (callNow) {
            func.apply(context, args);
          }
        };
      }
    }, {
      key: 'subscribeQuery',
      value: function subscribeQuery(query, callbacks) {
        var isNewSubscription = false;
        if (typeof this.subscriptions.queries[query] === 'undefined') {
          this.subscriptions.queries[query] = [];
          isNewSubscription = true;
        }
        var isCountSubscribed = isNewSubscription || this._isCountSubscribedQuery(query);
        this.subscriptions.queries[query].push(callbacks);
        if (this.connection.readyState === WebSocket.OPEN) {
          if (isNewSubscription) {
            this._subscribeQuery(query, typeof callbacks[2] !== 'undefined');
          } else if (!isCountSubscribed && callbacks[2] !== 'undefined') {
            this._unsubscribeQuery(query);
            this._subscribeQuery(query, true);
          }
        }
      }
    }, {
      key: 'subscribeUID',
      value: function subscribeUID(name, callbacks) {
        var isNewSubscription = false;
        if (typeof this.subscriptions.uids[name] === 'undefined') {
          this.subscriptions.uids[name] = [];
          isNewSubscription = true;
        }
        var isCountSubscribed = isNewSubscription || this._isCountSubscribedUID(name);
        this.subscriptions.uids[name].push(callbacks);
        if (this.connection.readyState === WebSocket.OPEN) {
          if (isNewSubscription) {
            this._subscribeUID(name, typeof callbacks[2] !== 'undefined');
          } else if (!isCountSubscribed && callbacks[2] !== 'undefined') {
            this._unsubscribeUID(name);
            this._subscribeUID(name, true);
          }
        }
      }
    }, {
      key: '_subscribeQuery',
      value: function _subscribeQuery(query, count) {
        this.connection.send(JSON.stringify({
          'action': 'subscribe',
          'query': query,
          'count': count
        }));
      }
    }, {
      key: '_subscribeUID',
      value: function _subscribeUID(name, count) {
        this.connection.send(JSON.stringify({
          'action': 'subscribe',
          'uid': name,
          'count': count
        }));
      }
    }, {
      key: '_isCountSubscribedQuery',
      value: function _isCountSubscribedQuery(query) {
        if (typeof this.subscriptions.queries[query] === 'undefined') {
          return false;
        }
        for (var callbacks = 0; callbacks < this.subscriptions.queries[query].length; callbacks++) {
          if (typeof this.subscriptions.queries[query][callbacks][2] !== 'undefined') {
            return true;
          }
        }
        return false;
      }
    }, {
      key: '_isCountSubscribedUID',
      value: function _isCountSubscribedUID(name) {
        if (typeof this.subscriptions.uids[name] === 'undefined') {
          return false;
        }
        for (var callbacks = 0; callbacks < this.subscriptions.uids[name].length; callbacks++) {
          if (typeof this.subscriptions.uids[name][callbacks][2] !== 'undefined') {
            return true;
          }
        }
        return false;
      }
    }, {
      key: 'unsubscribeQuery',
      value: function unsubscribeQuery(query, callbacks) {
        if (typeof this.subscriptions.queries[query] === 'undefined') {
          return;
        }
        var idx = this.subscriptions.queries[query].indexOf(callbacks);
        if (idx === -1) {
          return;
        }
        this.subscriptions.queries[query].splice(idx, 1);
        if (!this.subscriptions.queries[query].length) {
          delete this.subscriptions.queries[query];
          if (this.connection.readyState === WebSocket.OPEN) {
            this._unsubscribeQuery(query);
          }
        }
      }
    }, {
      key: 'unsubscribeUID',
      value: function unsubscribeUID(name, callbacks) {
        if (typeof this.subscriptions.uids[name] === 'undefined') {
          return;
        }
        var idx = this.subscriptions.uids[name].indexOf(callbacks);
        if (idx === -1) {
          return;
        }
        this.subscriptions.uids[name].splice(idx, 1);
        if (!this.subscriptions.uids[name].length) {
          delete this.subscriptions.uids[name];
          if (this.connection.readyState === WebSocket.OPEN) {
            this._unsubscribeUID(name);
          }
        }
      }
    }, {
      key: '_unsubscribeQuery',
      value: function _unsubscribeQuery(query) {
        this.connection.send(JSON.stringify({
          'action': 'unsubscribe',
          'query': query
        }));
      }
    }, {
      key: '_unsubscribeUID',
      value: function _unsubscribeUID(name) {
        this.connection.send(JSON.stringify({
          'action': 'unsubscribe',
          'uid': name
        }));
      }
    }, {
      key: 'updateArray',
      value: function updateArray(oldArr, update) {
        if (Array.isArray(update)) {
          var newArr = [].concat(_toConsumableArray(update));

          if (oldArr.length === 0) {
            // This will happen on the first update from a subscribe.
            oldArr.splice.apply(oldArr, [0, 0].concat(_toConsumableArray(newArr)));
            return;
          }

          var idMap = {};
          for (var i = 0; i < newArr.length; i++) {
            if (newArr[i] instanceof _Nymph.Nymph.getEntityClass('Nymph\\Entity') && newArr[i].guid) {
              idMap[newArr[i].guid] = i;
            }
          }
          var remove = [];
          for (var k in oldArr) {
            if (k <= 4294967294 && /^0$|^[1-9]\d*$/.test(k) && oldArr.hasOwnProperty(k)) {
              // This handles sparse arrays.
              k = Number(k);
              if (typeof idMap[oldArr[k].guid] === 'undefined') {
                // It was deleted.
                remove.push(k);
              } else if (newArr[idMap[oldArr[k].guid]].mdate !== oldArr[k].mdate) {
                // It was modified.
                oldArr[k].init(newArr[idMap[oldArr[k].guid]].toJSON());
                delete idMap[oldArr[k].guid];
              } else {
                // Item wasn't modified.
                delete idMap[oldArr[k].guid];
              }
            }
          }
          // Now we must remove the deleted ones.
          remove.sort(function (a, b) {
            // Sort backwards so we can remove in reverse order. (Preserves indices.)
            if (a > b) return -1;
            if (a < b) return 1;
            return 0;
          });
          for (var n = 0; n < remove.length; n++) {
            oldArr.splice(remove[n], 1);
          }
          // And add the new ones.
          for (var v in idMap) {
            if (idMap.hasOwnProperty(v)) {
              oldArr.splice(oldArr.length, 0, newArr[idMap[v]]);
            }
          }
        } else if (update != null && update.hasOwnProperty('query')) {
          var query = JSON.parse(update.query);

          if (update.hasOwnProperty('removed')) {
            for (var _i3 = 0; _i3 < oldArr.length; _i3++) {
              if (oldArr[_i3] != null && oldArr[_i3].guid === update.removed) {
                oldArr.splice(_i3, 1);
                return;
              }
            }
          }

          // Get the entity.
          var entity = void 0;
          if (update.hasOwnProperty('added')) {
            // A new entity.
            entity = _Nymph.Nymph.initEntity(update.data);
          }
          if (update.hasOwnProperty('updated')) {
            // Extract it from the array.
            for (var _i4 = 0; _i4 < oldArr.length; _i4++) {
              if (oldArr[_i4] != null && oldArr[_i4].guid === update.updated) {
                entity = oldArr.splice(_i4, 1)[0].init(update.data);
              }
            }
          }

          if (entity != null) {
            // Insert the entity in order.
            var sort = query[0].hasOwnProperty('sort') ? query[0].sort : 'cdate';
            var reverse = query[0].hasOwnProperty('reverse') ? query[0].reverse : false;
            var _i5 = void 0;

            if (reverse) {
              for (_i5 = oldArr.length; (oldArr[_i5] || {})[sort] < entity[sort] && _i5 > 0; _i5--) {}
            } else {
              for (_i5 = 0; (oldArr[_i5] || {})[sort] < entity[sort] && _i5 < oldArr.length; _i5++) {}
            }

            oldArr.splice(_i5, 0, entity);
          }
        }
      }
    }, {
      key: 'on',
      value: function on(event, callback) {
        if (!this.hasOwnProperty(event + 'Callbacks')) {
          return false;
        }
        this[event + 'Callbacks'].push(callback);
        return true;
      }
    }, {
      key: 'off',
      value: function off(event, callback) {
        if (!this.hasOwnProperty(event + 'Callbacks')) {
          return false;
        }
        var i = this[event + 'Callbacks'].indexOf(callback);
        if (i > -1) {
          this[event + 'Callbacks'].splice(i, 1);
        }
        return true;
      }
    }, {
      key: 'setToken',
      value: function setToken(token) {
        authToken = token;
        if (this.connection.readyState === WebSocket.OPEN) {
          this.connection.send(JSON.stringify({
            'action': 'authenticate',
            'token': authToken
          }));
        }
      }
    }]);

    return PubSub;
  }();

  // === Static Properties ===
  PubSub.connection = null;
  PubSub.pubsubURL = null;
  PubSub.rateLimit = null;
  PubSub.debouncers = {};
  PubSub.subscriptions = {
    queries: {},
    uids: {}
  };
  PubSub.connectCallbacks = [];
  PubSub.disconnectCallbacks = [];

  var PubSubSubscription = exports.PubSubSubscription = function PubSubSubscription(query, callbacks, unsubscribe) {
    _classCallCheck(this, PubSubSubscription);

    // === Instance Properties ===

    this.query = query;
    this.callbacks = callbacks;
    this.unsubscribe = unsubscribe;
  };

  if (typeof window !== 'undefined' && typeof window.NymphOptions !== 'undefined' && typeof window.NymphOptions.pubsubURL !== 'undefined') {
    PubSub.init(window.NymphOptions);
  }

  exports.default = PubSub;
});